@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2013-2019 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
+ *  Copyright (C) 2013-2021 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
  *  Copyright (C) 2008-2013 Sourcefire, Inc.
  *
  *  Author: aCaB <acab@clamav.net>
@@ -34,10 +34,13 @@
 
 #include <libmilter/mfapi.h>
 
-#include "libclamav/clamav.h"
-#include "shared/optparser.h"
-#include "shared/output.h"
-#include "libclamav/others.h"
+// libclamav
+#include "clamav.h"
+#include "others.h"
+
+// shared
+#include "optparser.h"
+#include "output.h"
 
 #include "connpool.h"
 #include "netcode.h"
@@ -774,6 +777,6 @@ sfsistat clamfi_envrcpt(SMFICTX *ctx, char **argv)
  * mode: c
  * c-basic-offset: 4
  * tab-width: 8
- * End: 
- * vim: set cindent smartindent autoindent softtabstop=4 shiftwidth=4 tabstop=8: 
+ * End:
+ * vim: set cindent smartindent autoindent softtabstop=4 shiftwidth=4 tabstop=8:
  */
